/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Faults definitions taken from RMP 440 LE User Manual pp. 73-78 */

#define _GNU_SOURCE
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

struct rmp_fault {
	uint32_t value;
	const char *msg;
};

/* Critical Faults */
static struct rmp_fault rmp_fault1[] = {
	{ 0x00000001, "ARCHITECT_FAULT_SPI_RECEIVE" },
	{ 0x00000002, "ARCHITECT_FAULT_SPI_TRANSMIT" },
	{ 0x00000004, "ARCHITECT_FAULT_SPI_RECEIVE_OVERRUN" },
	{ 0x00000008, "ARCHITECT_FAULT_SPI_RX_BUFFER_OVERRUN" },
	{ 0x00000010, "ARCHITECT_FAULT_COMMANDED_SAFETY_SHUTDOWN" },
	{ 0x00000020, "ARCHITECT_FAULT_COMMANDED_DISABLE" },
	{ 0x00000040, "ARCHITECT_FAULT_KILL_SWITCH_ACTIVE" },
	{ 0x00000080, "ARCHITECT_FAULT_FRAM_CONFIG_INIT_FAILED" },
	{ 0x00000100, "ARCHITECT_FAULT_FRAM_CONFIG_SET_FAILED" },
	{ 0x00000200, "ARCHITECT_FAULT_BAD_MODEL_IDENTIFIER" },
	{ 0x00000400, "ARCHITECT_FAULT_BAD_CCU_HW_REV" },
	{ 0x00000800, "ARCHITECT_FAULT_DECEL_SWITCH_ACTIVE" },
	{ 0x00001000, "CRITICAL_FAULT_INIT" },
	{ 0x00002000, "CRITICAL_FAULT_INIT_UIP_COMM" },
	{ 0x00004000, "CRITICAL_FAULT_INIT_PROPULSION" },
	{ 0x00008000, "CRITICAL_FAULT_INIT_TIMEOUT" },
	{ 0x00010000, "CRITICAL_FAULT_FORW_SPEED_LIMITER_HAZARD" },
	{ 0x00020000, "CRITICAL_FAULT_AFT_SPEED_LIMITER_HAZARD" },
	{ 0x00040000, "CRITICAL_FAULT_CHECK_STARTUP" },
	{ 0x00080000, "CRITICAL_FAULT_APP_VELOCITY_CTL_FAILED" },
	{ 0x00100000, "CRITICAL_FAULT_APP_POSITION_CTL_FAILED" },
	{ 0x00200000, "CRITICAL_FAULT_ABB_SHUTDOWN" },
	{ 0x00400000, "CRITICAL_FAULT_AP_MODE_TRANS_TIMEOUT" },
	{ 0x00800000, "CRITICAL_FAULT_PITCH_ANGLE_EXCEEDED" },
	{ 0x01000000, "CRITICAL_FAULT_ROLL_ANGLE_EXCEEDED" },
};

static struct rmp_fault rmp_fault2[] = {
	{ 0x00000001, "COMM_FAULT_UIP_MISSING_UIP_DATA" },
	{ 0x00000002, "COMM_FAULT_UIP_UNKNOWN_MESSAGE_RECEIVED" },
	{ 0x00000004, "COMM_FAULT_UIP_BAD_CHECKSUM" },
	{ 0x00000008, "COMM_FAULT_UIP_TRANSMIT" },
	{ 0x00000010, "COMM_FAULT_UI_BAD_MOTION_CMD" },
	{ 0x00000020, "COMM_FAULT_UI_UNKOWN_CMD" },
	{ 0x00000040, "COMM_FAULT_UI_BAD_PACKET_CHECKSUM" },
	{ 0x00010000, "INTERNAL_FAULT_HIT_DEFAULT_CONDITION" },
	{ 0x00020000, "INTERNAL_FAULT_HIT_SPECIAL_CASE"},
};

static struct rmp_fault rmp_fault3[] = {
	{ 0x00000001, "SENSOR_FAULT_2P5V_VREF_RANGE_FAULT" },
	{ 0x00000002, "SENSOR_FAULT_7P2V_VBAT_RANGE_FAULT" },
	{ 0x00000004, "SENSOR_FAULT_7P2V_VBAT_WARNING" },
	{ 0x00000008, "SENSOR_FAULT_7P2V_BATT_INBALANCE_FAULT" },
	{ 0x00000010, "SENSOR_FAULT_7P2V_BATT_TEMPERATURE_FAULT" },
	{ 0x00000020, "SENSOR_FAULT_DIGITAL_INPUT" },
	{ 0x00000040, "SENSOR_FAULT_RANGE" },
	{ 0x00000080, "SENSOR_FAULT_DEFAULT" },
	{ 0x00000100, "SENSOR_FAULT_5V_MONITOR_RANGE_FAULT" },
	{ 0x00000200, "SENSOR_FAULT_12V_MONITOR_RANGE_FAULT" },
	{ 0x00010000, "BSA_FAULT_SIDE_A_MISSING_BSA_DATA" },
	{ 0x00020000, "BSA_FAULT_SIDE_B_MISSING_BSA_DATA" },
	{ 0x00040000, "BSA_FAULT_UNKNOWN_MESSAGE_RECEIVED" },
	{ 0x00080000, "BSA_FAULT_TRANSMIT_A_FAILED" },
	{ 0x00100000, "BSA_FAULT_TRANSMIT_B_FAILED" },
	{ 0x00200000, "BSA_FAULT_DEFAULT" },
	{ 0x00400000, "BSA_FAULT_SIDE_A_RATE_SENSOR_SATURATED" },
	{ 0x00800000, "BSA_FAULT_SIDE_B_RATE_SENSOR_SATURATED" },
	{ 0x01000000, "BSA_FAULT_SIDE_A_TILT_SENSOR_SATURATED" },
	{ 0x02000000, "BSA_FAULT_SIDE_B_TILT_SENSOR_SATURATED" },
	{ 0x04000000, "PSE_FAULT_COMPARISON" },
};

static struct rmp_fault rmp_fault4[] = {
	{ 0x00000001, "MCU_FAULT_MCU_0_IS_DEGRADED" },
	{ 0x00000002, "MCU_FAULT_MCU_0_IS_FAILED" },
	{ 0x00000004, "MCU_FAULT_MCU_0_REQUESTS_REDUCED_PERFORMANCE" },
	{ 0x00000008, "MCU_FAULT_MCU_0_REQUESTS_ZERO_SPEED" },
	{ 0x00000010, "MCU_FAULT_MCU_1_IS_DEGRADED" },
	{ 0x00000020, "MCU_FAULT_MCU_1_IS_FAILED" },
	{ 0x00000040, "MCU_FAULT_MCU_1_REQUESTS_REDUCED_PERFORMANCE" },
	{ 0x00000080, "MCU_FAULT_MCU_1_REQUESTS_ZERO_SPEED" },
	{ 0x00000100, "MCU_FAULT_MCU_2_IS_DEGRADED" },
	{ 0x00000200, "MCU_FAULT_MCU_2_IS_FAILED" },
	{ 0x00000400, "MCU_FAULT_MCU_2_REQUESTS_REDUCED_PERFORMANCE" },
	{ 0x00000800, "MCU_FAULT_MCU_2_REQUESTS_ZERO_SPEED" },
	{ 0x00001000, "MCU_FAULT_MCU_3_IS_DEGRADED" },
	{ 0x00002000, "MCU_FAULT_MCU_3_IS_FAILED" },
	{ 0x00004000, "MCU_FAULT_MCU_3_REQUESTS_REDUCED_PERFORMANCE" },
	{ 0x00008000, "MCU_FAULT_MCU_3_REQUESTS_ZERO_SPEED" },
	{ 0x00010000, "MCU_FAULT_MISSING_MCU_0_DATA" },
	{ 0x00020000, "MCU_FAULT_MISSING_MCU_1_DATA" },
	{ 0x00040000, "MCU_FAULT_MISSING_MCU_2_DATA" },
	{ 0x00080000, "MCU_FAULT_MISSING_MCU_3_DATA" },
	{ 0x00100000, "MCU_FAULT_UNKNOWN_MESSAGE_RECEIVED" },
};

static struct rmp_fault *rmp_faults[] = {
	rmp_fault1,
	rmp_fault2,
	rmp_fault3,
	rmp_fault4
};

#ifdef notyet
static struct rmp_fault mcu_specific_faults[] = {
	{ 0x00000001, "MCU_TRANS_BATTERY_TEMP_WARNING" },
	{ 0x00000002, "MCU_TRANS_BATTERY_COLD_REGEN" },
	{ 0x00000004, "MCU_UNKNOWN" },
	{ 0x00000008, "MCU_UNKNOWN" },
	{ 0x00000010, "MCU_TRANS_LOW_BATTERY" },
	{ 0x00000020, "MCU_TRANS_BATT_OVERVOLTAGE" },
	{ 0x00000040, "MCU_CRITICAL_BATT_OVERVOLTAGE" },
	{ 0x00000080, "MCU_CRITICAL_EMPTY_BATTERY" },
	{ 0x00000100, "MCU_CRITICAL_BATTERY_TEMP" },
	{ 0x00000200, "MCU_COMM_CU_BCU_LINK_DOWN" },
	{ 0x00000400, "MCU_COMM_INITIALIZATION_FAILED" },
	{ 0x00000800, "MCU_COMM_FAILED_CAL_EEPROM" },
	{ 0x00001000, "MCU_POWER_SUPPLY_TRANSIENT_FAULT" },
	{ 0x00002000, "MCU_POWER_SUPPLY_12V_FAULT" },
	{ 0x00004000, "MCU_POWER_SUPPLY_5V_FAULT" },
	{ 0x00008000, "MCU_POWER_SUPPLY_3V_FAULT" },
	{ 0x00010000, "MCU_JUNCTION_TEMP_FAULT" },
	{ 0x00020000, "MCU_MOTOR_WINDING_TEMP_FAULT" },
	{ 0x00040000, "MCU_MOTOR_DRIVE_FAULT" },
	{ 0x00080000, "MCU_MOTOR_DRIVE_HALL_FAULT" },
	{ 0x00100000, "MCU_MOTOR_DRIVE_AMP_FAULT" },
	{ 0x00200000, "MCU_MOTOR_DRIVE_AMP_ENABLE_FAULT" },
	{ 0x00400000, "MCU_MOTOR_DRIVE_AMP_OVERCURRENT_FAULT" },
	{ 0x00800000, "MCU_MOTOR_DRIVE_VOLTAGE_FEEDBACK_FAULT" },
	{ 0x01000000, "MCU_FRAME_FAULT" },
	{ 0x02000000, "MCU_BATTERY_FAULT" },
	{ 0x08000000, "MCU_MOTOR_STUCK_RELAY_FAULT" },
	{ 0x10000000, "MCU_ACTUATOR_POWER_CONSISTENCY_FAULT" },
	{ 0x20000000, "MCU_ACTUATOR_HALT_PROCESSOR_FAULT" },
	{ 0x40000000, "MCU_ACTUATOR_DEGRADED_FAULT"},
};
#endif

/*----------------------------------------------------------------------*/

/**
 ** Construction incrementale de chaine
 **/
void
bufcat(char **buf, char *fmt, ...)
{
    va_list ap;
    char *r = NULL;
    int s;

    va_start(ap, fmt);
    s = vasprintf(&r, fmt, ap);
    va_end(ap);
    if (s == -1) {
	fprintf(stderr, "vasprintf failed\n");
	_exit(2);
    }
    if (*buf != NULL) {
	s = asprintf(buf, "%s%s", *buf, r);
	free(r);
    } else {
	*buf = r;
    }
} /* bufcat */


char *
rmp_decode_fault(uint32_t *regs)
{
	int i, j;
	uint32_t bit;
	char *result = NULL;
	int first = 1;

	for (i = 0; i < 3; i++) {
		for (j = 0; j < 32; j++) {
			bit = 1<<j;
			if ((regs[i] & bit) == bit) {
				bufcat(&result, first ? "%s" : "|%s\n", 
				    rmp_faults[i][bit]);
				first = 0;
			}
		}
	}
	return result;
}
